#!/usr/bin/env python
from datetime import date, datetime
from mqtt_client import MqttClient
import json

class Logger:

    # Vars
    f = None # File handler (opened once log file is started)
    header_list = None # Populated with each column name
    WANDER_HEADERS = ["schedule_active", "battery", "loiter_time_remaining", "signal_strength"] # Column names only present in small wander messages

    @staticmethod
    def message_handler(message):
        """
        This method is called once for each message received after it's added as a handler to the mqtt_client instance
        """
        # Generate file when first message is received
        if not Logger.f:
            is_wander_msg = Logger.create_header_list(message)
            # Don't start log w/ wander message since they're missing most keys from standard messages
            if is_wander_msg:
                return
            Logger.create_log_file()

        # Parse incoming json into dict
        payload = str(message.payload)
        payload_dict = json.loads(payload)
        
        # Find each column name in the message and put the value, or if not there insert a blank space
        for i, key in enumerate(Logger.header_list):
            if key != "topic": # topic isn't a key in the message payload, so it must be accessed differently
                if key in payload_dict: # Only certain that a given message has the key 'date'
                    if i == (len(Logger.header_list) - 1): # Final column should end with a newline, not a trailing tab
                        Logger.f.write("%s" % str(payload_dict[key]))
                    else:
                        Logger.f.write("%s\t" % str(payload_dict[key]))
                else:
                    Logger.f.write("\t")
            else:
                Logger.f.write("%s\t" % message.topic)

        Logger.f.write("\n")
        Logger.f.flush()

    @staticmethod
    def create_header_list(message):
        """
        This generates the list of column names to insert on line 1 of the log file based on a message
        """
        # Don't set headers based on small message
        if "wander" in message.topic:
            return 1
        payload = str(message.payload)
        payload_dict = json.loads(payload)
        Logger.header_list = list(payload_dict.keys())


        # Add topic column, force date to the left, append additional column headers
        Logger.header_list.insert(0,"topic") # Current list: ['topic', ..., 'date', ...]
        Logger.header_list.remove("date") # Current list: ['topic', ...]
        Logger.header_list.insert(0,"date") # Current list: ['date', 'topic', ...]
        # Now the left side of the log matches what is seen the the GUI

        # Adds the wander message columns on the far right of the log so they're grouped
        for i in Logger.WANDER_HEADERS:
            Logger.header_list.append(i)

        return None

    @staticmethod
    def create_header_line(filename):
        """
        Writes the first line to the log file with all column names in order
        """
        # Add first line to the log file
        for i, val in enumerate(Logger.header_list):
            if i == (len(Logger.header_list) - 1):
                Logger.f.write("%s\n" % val)
            else:
                Logger.f.write("%s\t" % val)

    @staticmethod
    def create_log_file():
        """
        Creates the log file and opens it for writing
        """
        # Generate file with timestamp name
        curr_date = str(date.today())
        dt = datetime.now()
        time = dt.strftime("%Hh-%Mm-%Ss")

        filename = "wander-" + curr_date + "-" + time + ".tsv"
        Logger.f = open(filename, "w+")
        Logger.create_header_line(filename)

    @staticmethod
    def start():
        """
        Adds the message handler to the mqtt_client so message_handler() is called once per message received
        """
        MqttClient.add_handler(Logger.message_handler)

    @staticmethod
    def stop():
        # Must be called to ensure all log lines are written
        try:
            Logger.f.close()
        except Exception:
            pass

if __name__ == "__main__":
    Logger.message_handler("test")
