#!/usr/bin/env python
import paho.mqtt.client as mqtt
import json

class MqttClient:

    # Vars
    BROKER_IP = "172.16.10.50" # Default broker IP is only available on IWC network
    BROKER_PORT = 1883
    SUBSCRIPTION_TOPIC = "#"
    USERNAME = None
    PASSWORD = None
    client = None
    handler_functions = []
    connected = False

    # Called each time the client makes a connection
    def on_connect(client, userdata, flags, rc):
        MqttClient.connected = True

    # Called each time the client loses/ends a connection
    def on_disconnect(client, userdata, rc):
        MqttClient.connected = False

    @staticmethod
    def subscribe(topic):
        MqttClient.client.subscribe(topic, qos=1)

    @staticmethod
    def connect_to_broker(ip=BROKER_IP, port=BROKER_PORT, username=None, password=None):
        # Changes broker while running (called from GUI)
        MqttClient.BROKER_IP = ip
        MqttClient.BROKER_PORT = int(port)
        MqttClient.USERNAME = username
        MqttClient.PASSWORD = password
        if MqttClient.client:
            MqttClient.client.loop_stop() # Client must be shut down and re-created (can't change IP for client once initialized)
        MqttClient.run()

    @staticmethod
    def create_client():
        MqttClient.client = mqtt.Client()
        return MqttClient.client

    @staticmethod
    def connect(ip=BROKER_IP, port=BROKER_PORT):
        MqttClient.client = MqttClient.create_client()

        # Configure username/password if present
        if MqttClient.USERNAME:
            MqttClient.client.username_pw_set(username=MqttClient.USERNAME, password=MqttClient.PASSWORD)

        connection = MqttClient.client.connect(ip, port)

        if connection == 0:
            return True
        else:
            return False

    def on_message(self, userdata, message):
        """
        This is the callback function connected to paho, it is called each time a message is received 

        It's the only non-static method in the class. If made static it will break the 2nd time run is called (when the broker ip/port is changed on the fly)
        """
        print("RECEIVED:", str(message.payload.decode("utf-8")))
        print("TOPIC:", str(message.topic))
        print("------------------------", flush=True)
        
        # Must be decoded as utf-8 the first time, after this there's no need to decode, just cast the payload to a string
        payload = str(message.payload.decode("utf-8"))
        payload_dict = json.loads(payload)
        message.payload = json.dumps(payload_dict)

        # Executes each stored handler function each time a message is executed
        for func in MqttClient.handler_functions:
            func(message)

    # Each handler added is called once with each received message
    @staticmethod
    def add_handler(function):
        MqttClient.handler_functions.append(function)

    @staticmethod
    def run():
        MqttClient.create_client()
        # try/except catches issues w/ invalid IPs or failure to connect to the broker
        try:
            if MqttClient.connect(MqttClient.BROKER_IP, MqttClient.BROKER_PORT):
                MqttClient.client.on_message=MqttClient.on_message
                MqttClient.client.on_connect = MqttClient.on_connect
                MqttClient.client.on_disconnect = MqttClient.on_disconnect
                MqttClient.client.loop_start()
                MqttClient.subscribe(MqttClient.SUBSCRIPTION_TOPIC)
        except Exception as ex:
            print(ex, flush=True)
            raise ex
            

if __name__ == "__main__":
    MqttClient.run()
    while True:
        pass
