from ui_view import UIView, MqttConnectDialog
from tkinter import END, DISABLED, NORMAL, StringVar
from enum import Enum
from mqtt_client import MqttClient
from rest_client import RequestTypes, RestClient
import json
import uuid



class UIController():
    """
    A class for communicating with the backend clients and 
    passing data to the UIView to be displayed.
    """

    # A dictionary of all of mqtt messages that have been heard
    messages_by_id = {}
    # Lines of mqtt messages that have been heard but not displayed to the user yet
    lines_to_insert = []
    # The ID of the last mqtt message that was heard
    latest_message_id = None
    # The ID of the user selected mqtt message
    selected_message_id = None
    # The mqtt filter to apply, must be more restrictive than door/+/event/#
    mqtt_filter = ""
    # Whether to pause the flow of incoming messages
    paused = False
    # The currently selected request type
    selected_request_type = None
    # Whether the application was started with arguments
    started_with_args = False

    def __init__(self, filter=None):
        if filter:
            self.mqtt_filter = filter
        else:
            self.mqtt_filter = MqttClient.SUBSCRIPTION_TOPIC

        self.view = UIView()
        self._register_mqtt_handlers()
        self._configure_gui()
        self._update_mqtt_message_list()
        self._update_message_status()
        self._check_for_disconnect()


    def mqtt_message_handler(self, mqtt_message):
        """
        Handler to be called by the MqttClient for each
        incoming message. It formats the message for showing 
        in the GUI and adds the full message to the internal dictionary

        mqtt_message -- the incoming message to handle
        """
        # If paused, don't handle incoming messages
        if self.paused:
            return

        # Assign a unique ID to keep track of messages
        message_id = str(uuid.uuid4())
        formatted_message = ""

        try:
            topic = str(mqtt_message.topic)
            # If the topic of the message doesn't match the filter, bail
            if not self._check_filter(topic):
                return
            
            date = self._decode_payload(mqtt_message.payload)['date']
            # The formatted message shows in the mqtt message list in the
            # GUI.  Due to the width of the list, the message_id is intentionally
            # hidden from view.  When the user clicks on a message in the list
            # we can use the hidden message_id to look up the full message for display
            formatted_message = date.ljust(30) + topic.ljust(100) + message_id

        except Exception as ex:
            self._update_app_status("Cannot handle incoming message!  Malformed message received?", error = True)
            formatted_message = "Malformed Message"
            print(ex)

        finally:
            if formatted_message != "":
                self.latest_message_id = message_id
                self.lines_to_insert.append(formatted_message)
                self.messages_by_id[message_id] = mqtt_message


    def show_live_status(self):
        """
        Clears user selection and goes back to showing the
        status of the most recent MQTT message in the GUI.  This 
        is called when the user clicks the "Live Status" button.
        """
        # Disable the live status button
        self.view.mqtt_live_status_button.config(state = DISABLED)
        # Clear any selected devices in the list
        self.view.mqtt_message_list.selection_clear(0, END)
        # Scroll to the top of the list
        self.view.mqtt_message_list.yview(0)
        self._update_app_status("Showing the live Message Status for incoming messages.")


    def set_filter(self, filter=None):
        """
        Sets the mqtt filter, this only applies to the GUI not the
        logger and does not change the MQTT client.
        """
        if filter:
            self.mqtt_filter = filter
        else:
            self.mqtt_filter = self.view.mqtt_filter_entry.get()
        self._update_app_status("MQTT filter applied successfully.")


    def connect_to_broker_address(self, args=""):
        """Sets the mqtt broker address and connects (with optional credentials), this will restart the client"""
        if self.check_args(args):
            address  = str(args.address)
            username = args.username
            password = args.password
            self.started_with_args = True

        else:
            # Show a connect dialog to the user
            dialog = MqttConnectDialog(self.view, title = "Connect To Broker", mqtt_broker_address = "{}:{}".format(MqttClient.BROKER_IP, MqttClient.BROKER_PORT))
            # When the dialog closes, if there is no result the dialog was cancelled
            if not dialog.result:
                return
            # Parse the fields out of the dialog
            address, username, password = dialog.result
            # Pause incoming messages
            self.paused = True

        try:
            ip, port = address.split(":")
            MqttClient.connect_to_broker(ip = ip, port = port, username = username, password = password)
            self._update_app_status("Connected to MQTT broker successfully")
            self._update_mqtt_connection_status(connected = True, address = "{}:{}".format(ip, port))
        except Exception:
            self._update_app_status("Unable to connect to broker, check parameters.", error=True)
            self._update_mqtt_connection_status(connected = False)
        finally:
            # Clear the list of messages and then unpause the incoming messages
            self._clear_messages()
            self.paused = False


    def check_args(self, args):
        """
        Check if the address was defined in the arguments and if so return it
        """
        if args:
            return args.address
        else:
            return False


    def start_application(self):
        """
        Tells the application to start, this is blocking so
        all configuration needs to be done prior to running this
        """
        if not self.started_with_args:
            self._update_app_status("Welcome to the Inovonics MQTT Client.  Click 'Connect To Broker' to begin!")
        self.view.mainloop()


    def send_request(self):
        """
        Send a request based on the fields that the user input
        """
        # Get the fields that the user entered
        base_url = self.view.request_base_url_entry.get()
        request_url = self.view.request_url_entry.get()
        request_body = self.view.request_body.get("1.0", END)
        # Set the base URL
        RestClient.set_base_url(base_url)
        # Make the call using the REST client
        response = None
        try:
            response = RestClient.make_request(self.selected_request_type.get(), request_url, body=request_body)
            # Get the portions of the response we need
            status = response.status_code
            try:
                body = response.json()
            except:
                body = response.text
            # Print out the response status
            if response.ok:
                self.view.response_status.config(text = "{} OK".format(status), fg = "green")
            else:
                self.view.response_status.config(text = "{} ERROR".format(status), fg="red")
        except Exception as ex:
            status = "???"
            body = ex
            self.view.response_status.config(text = "{} ERROR".format(status), fg="red")

        # Print out the response content
        self.view.response_body.delete("1.0", END)
        try:
            output = json.dumps(body, indent=4, sort_keys=True)
        except:
            output = body
        self.view.response_body.insert(END, output)


    def copy_response(self):
        """
        Copy the response field to the request field for the REST client
        """
        response_body = self.view.response_body.get("1.0", END)
        self.view.request_body.delete("1.0", END)
        self.view.request_body.insert(END, response_body)


    def _clear_messages(self):
        """Clear the messages in the GUI and in the dictionary"""
        self.paused = True
        self.messages_by_id = {}
        self.lines_to_insert = []
        self.latest_message_id = None
        self.selected_message_id = None
        # Clear the list of mqtt messages in the GUI
        self.view.mqtt_message_list.delete(0, END)
        # Clear message status view
        self.view.status_list_text.delete("1.0", END)
        self.paused = False


    def _update_app_status(self, string, error = False):
        """
        Update the application status label with text

        string -- the string to set as the status text
        error -- (optional) bool indicating whether the 
                  status text should be red or not
        """
        if error:
            self.view.application_status_label.config(text = "ERROR:  " + string, fg = "red")
        else:
            self.view.application_status_label.config(text = string, fg = "grey")


    def _check_filter(self, topic):
        """
        Checks whether the incoming message topic matches the filter

        topic -- the mqtt message topic to check against the filter
        returns True if the message matches, False if not
        """
        if self.mqtt_filter == "":
            return True

        filter_parts = self.mqtt_filter.split("/")
        topic_parts = topic.split("/")

        # Compare the parts of the filter against the parts of the incoming topic
        for index, filter_part in enumerate(filter_parts):
            # If the filter part is #, the rest of the topic will match so return True
            if filter_part == "#":
                return True
            # If the filter part does not match and the filter part is not +, return False
            if filter_part != topic_parts[index] and filter_part != "+":
                return False
        # If the loop completed without returning, it must match, return True
        return True


    def _register_mqtt_handlers(self):
        """Add handlers to the mqtt client"""
        MqttClient.add_handler(self.mqtt_message_handler)


    def _configure_gui(self):
        """Tie handlers to GUI components and set defaults"""
        # Configuring buttons to use controller methods when clicked.
        self.view.mqtt_live_status_button.config(command = self.show_live_status)
        self.view.mqtt_filter_submit_button.config(command = self.set_filter)
        self.view.mqtt_broker_address_submit_button.config(command = self.connect_to_broker_address)
        # Set the intial topic users see
        self.view.mqtt_filter_entry.insert(END, self.mqtt_filter)

        self._update_mqtt_connection_status(connected = False)

        self.selected_request_type = StringVar(self.view)

        for request_type in RequestTypes:
            self.view.add_request_type(request_type.name, self.selected_request_type)
        
        self.selected_request_type.set(RequestTypes.GET.name)
        self.view.request_base_url_entry.insert(END, RestClient.base_url)
        self.view.request_send_button.config(command = self.send_request)
        self.view.response_copy_button.config(command = self.copy_response)
        

    def _update_mqtt_message_list(self):
        """
        Every 500ms, check for new mqtt messages to display
        and add them to the list in the GUI.

        This method only needs to be called one time, and will
        continue to run for the life of the application.
        """
        while len(self.lines_to_insert) > 0:
            line = self.lines_to_insert.pop(0)
            self.view.mqtt_message_list.insert(0, line)
        self.view.after(500, self._update_mqtt_message_list)


    def _update_mqtt_connection_status(self, connected = False, address = ""):
        if connected:
            self.view.mqtt_connection_status_label.config(text = "Connected " + address, foreground = "green")
        else:
            self.view.mqtt_connection_status_label.config(text = "Disconnected", foreground = "red")


    def _check_for_disconnect(self):
        """
        Every 3000ms (3s), check if the broker is still connected.
        If it is not, alert the user.
        """
        # If we are not connected but are listening for new messages, the broker must have disconnected
        if not MqttClient.connected and not self.paused:
            self.paused = True
            self._update_mqtt_connection_status(connected = False)
            self._update_app_status("Broker connection dropped!", error = True)
        # If we are connected but not listening for new messages, we must have reconnected with the broker
        if MqttClient.connected and self.paused:
            self.paused = False
            self._update_mqtt_connection_status(connected = True, address = "{}:{}".format(MqttClient.BROKER_IP, MqttClient.BROKER_PORT))
            self._update_app_status("Reconnected to the MQTT broker.")
            # self.view.mqtt
        # Re-run this method every 3 seconds
        self.view.after(3000, self._check_for_disconnect)


    def _update_message_status(self):
        """
        Update the full message displayed in the GUI.  This runs
        every 500ms and checks if a different message needs to
        be displayed, and if so displays it.

        This method only needs to be called one time, and will
        continue to run for the life of the application.
        """
        message_id = None
        # Get the list of user selected lines (can only be zero or one based on the GUI config)
        selected_indexes = self.view.mqtt_message_list.curselection()
        user_selection = False
        # If there is a selected line, try to get the message id for it
        if len(selected_indexes) > 0:
            user_selection = True
            self.view.mqtt_live_status_button.config(state = NORMAL)
            selected_row = self.view.mqtt_message_list.get(selected_indexes[0])
            # Message id is the last part of the row, and is hidden from user view by padding
            message_id = selected_row.split()[2]
        # Otherwise just get the latest message id
        elif self.latest_message_id is not None:
            message_id = self.latest_message_id 
        # If there is a message id to display, and it does not match the currently
        # displayed message id, get the message from the dictionary and display it
        if message_id is not None and message_id != self.selected_message_id:
            if user_selection:
                self._update_app_status("Showing Message Status for user selected message.")
            self.view.status_list_text.delete("1.0", END)
            # Try printing out the message payload
            try:
                message = self._decode_payload(self.messages_by_id[message_id].payload)
                self._print_dictionary(message, 0)
                self.selected_message_id = message_id
            except Exception as ex:
                self._update_app_status("Cannot display Message Status! Malformed payload?", error = True)
                print(ex)

        self.view.after(500, self._update_message_status)

    
    def _decode_payload(self, payload):
        """
        Decode mqtt payload and convert it to JSON

        payload -- the mqtt payload to convert
        """
        return json.loads(str(payload))


    def _print_dictionary(self, dictionary, padding):
        """
        Print a python dictionary in human readable form.

        dictionary -- the dictionary to print
        padding -- the amount of space padding to put before each line
        """
        for key, value in dictionary.items():
            self.view.status_list_text.insert(END, " " * padding + (str(key) + ": ").ljust(30))
            value_is_dict_list = False
            # If the value is a list of dictionaries, recursively call this 
            # method with increased padding to print out the sub dictionaries
            if isinstance(value, list) and len(value) > 0:
                if isinstance(value[0], dict):
                    value_is_dict_list = True
                    for item in value:
                        self.view.status_list_text.insert(END, '\n')
                        self._print_dictionary(item, 3)
            if not value_is_dict_list:
                self.view.status_list_text.insert(END, str(value) + '\n')
